#!/usr/bin/python
"""Semantic Web API"""

__author__ = 'Sean B. Palmer'
__license__ = 'Copyright (C) 2002 Sean B. Palmer. GNU GPL 2'

import UserList, copy

SUBJECT, PREDICATE, OBJECT, FORMULA = 0, 1, 2, 3

class Article: 
   """This is the main kind of "thing" in the API."""
   def __init__(self, s): 
      s, self.__r = str(s), str(s)
      if s[0] == '"': self.value, self.type = s[1:-1], 'Literal'
      elif s[0] == '?': self.value, self.type = s[1:], 'Univar'
      elif s[0] == '_': self.value, self.type = s[2:], 'Exivar'
      elif s[0] == '<': self.value, self.type = s[1:-1], 'URI'
      else: self.value, self.type, self.__r = s, 'URI', '<%s>' % s
   def __repr__(self): return self.__r
   def set(self, s): self.__init__(s)

class Tuple(tuple): 
   """This is basically an n-ary association of Articles."""
   def __eq__(self, x):
      if not isinstance(x, Tuple): return 0
      elif not (len(self.data) == len(x.data)): return 0
      else: return `self.data` == `x.data`

class Triple(Tuple): 
   def __init__(self, data): 
      if len(data) == 3: tuple.__init__(self, data)
      else: raise "Input of wrong length " + `len(data)`

   def __repr__(self): 
      return 'Triple'+Tuple.__repr__(self)

   def __eq__(self, x):
      if not isinstance(x, Triple): return 0
      elif not (len(x) == 3): return 0
      else: return `self` == `x`

class TupleStore(list): 
   """This is a kind of list of tuples."""
   def __init__(self, data=None): 
      list.__init__(self)
      self._tuples = []
      self._vars = []
      self._counter = 1
      if data: self.extend(data)

   def feed(self, store): 
      varmap = {}
      for t in store: 
         t = t[:]
         for i in range(len(t)): 
            if t[i].type[-3:] == 'var': 
               ti = `t[i]`
               if ti in varmap.keys(): t[i] = varmap[ti]
               else: 
                  while `t[i]` in self._vars: 
                     t[i] = Article(`t[i]` + `self._counter`)
                     self._counter += 1
                  varmap[ti] = t[i]
         self.append(t)

   def query(self, q, loose=1): 
      r = []
      for t in self: # Query the triple against the store
         if len(t) != len(q): pass
         else: 
            add = 1
            for i in range(len(t)): 
               if loose: 
                  if (q[i].type[-3:] != 'var') and (`t[i]` != `q[i]`): 
                     add = 0
               else: 
                  if (q[i].type != 'Univar') and (`t[i]` != `q[i]`): 
                     add = 0
            if add: r.append(t)
      return r

   def remove(self, t): 
      q = self.query(t)
      return TupleStore(filter(lambda x: (x not in q), self))

class TripleStore(TupleStore): 
   """This is a kind of list of tuples."""
   def __init__(self, data=None): 
      list.__init__(self)
      self._triples = {}
      self._tuples = []
      self._vars = []
      self._counter = 1
      if data: self.extend(data)

   def _triplesAppend(self, (s, p, o)): 
      s, p, o = (`s`, `p`, `o`)
      if s in self._triples.keys(): 
         if p in self._triples[s]: self._triples[s][p] += [o]
         else: self._triples[s][p] = [o]
      else: self._triples[s] = { p: [o] }

   def append(self, t): 
      """Append the triple t to the store if it does not already 
         contain it."""
      if not isinstance(t, Triple): raise `t`+' is not a Triple'

      for b in t: 
         if b.type[-3:] == 'var': self._vars.append(`b`)

      if `t[0]` not in self._triples.keys(): 
         self._triplesAppend(t)
         TupleStore.append(self, t)
      elif `t[1]` not in self._triples[`t[0]`].keys(): 
         self._triplesAppend(t)
         TupleStore.append(self, t)
      elif `t[2]` not in self._triples[`t[0]`][`t[1]`]: 
         self._triplesAppend(t)
         TupleStore.append(self, t)
      else: pass

   def extend(self, *triplelists): 
      for triplelist in triplelists: 
         for triple in triplelist: self.append(triple)

   def remove(self, t): 
      q = self.query(t)
      return TripleStore(filter(lambda x: (x not in q), self))

class Store(dict): 
   def __len__(self): 
      result = 0
      for formula in self.formulae(): result += len(self[formula])
      return result

   def formulae(self): return self.keys()

   def append(self, triple, formula): 
      if `formula` in self.formulae(): self[`formula`].append(triple)
      elif formula is None: 
         if None in self.formulae(): self[None].append(triple)
         else: self[None] = TripleStore([triple])
      else: self[`formula`] = TripleStore([triple])

if __name__=="__main__": 
   print __doc__